package net.sergeych.crypto

import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient

/**
 * Multi-signed data box. Use [SignedBox.invoke] to easily create
 * instances and [SignedBox.plus] to add more signatures (signing keys), and
 * [SignedBox.contains] to check for a specific key signature presence.
 *
 * It is serializable and checks integrity on deserialization. If any of seals does not
 * match the signed [message], it throws [IllegalSignatureException] _on deserialization_.
 * E.g., if you have it deserialized, it is ok, check it contains all needed keys among
 * signers.
 *
 * __The main constructor is used for deserializing only__. Don't use it directly unless you
 * know what you are doing as it may be dangerous.Use one of the above to create or change it.
 */
@Serializable
class SignedBox(
    val message: UByteArray,
    private val seals: List<Seal>,
    @Transient
    private val checkOnInit: Boolean = true
) {

    /**
     * If this instance is not signed by a given key, return new instance signed also by this
     * key, or return unchanged (same) object if it is already signed by this key; you
     * _can't assume it always returns a copied object!_
     */
    operator fun plus(key: Key.Signing): SignedBox =
        if (key.verifying in this) this
        else SignedBox(message, seals + Seal.create(key, message), false)

    /**
     * Check that it is signed with a specified key.
     */
    operator fun contains(verifyingKey: Key.Verifying): Boolean {
        return seals.any { it.key == verifyingKey }
    }

    init {
        if (seals.isEmpty()) throw IllegalArgumentException("there should be at least one seal")
        if (checkOnInit) {
            if (!seals.all { it.verify(message) }) throw IllegalSignatureException()
        }
    }

    /**
     * A key + signature pair for [SignedBox] boxes, usually you don't use it
     * directly bug call [SignedBox] constructor or [SignedBox.plus] to
     * add seals.
     */
    @Serializable
    data class Seal(val key: Key.Verifying, val signature: UByteArray) {

        fun verify(message: UByteArray): Boolean = key.verify(signature, message)

        companion object {
            fun create(key: Key.Signing, message: UByteArray): Seal {
                return Seal(key.verifying, key.sign(message))
            }
        }
    }

    companion object {
        /**
         * Create a new instance with a specific data sealed by one or more
         * keys. At least one key is required to disallow providing not-signed
         * instances, e.g. [SignedBox] is guaranteed to be properly sealed when
         * successfully instantiated.
         *
         * @param data a message to sign
         * @param keys a list of keys to sign with, should be at least one key.
         * @throws IllegalArgumentException if keys are not specified.
         */
        operator fun invoke(data: UByteArray, vararg keys: Key.Signing): SignedBox =
            SignedBox(data, keys.map { Seal.create(it, data) }, false)
    }
}