package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.*
import net.sergeych.crypto.Key
import net.sergeych.kiloparsec.KiloClientConnection
import net.sergeych.kiloparsec.KiloInterface
import net.sergeych.kiloparsec.ProxyDevice
import net.sergeych.mp_logger.*
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter
import net.sergeych.utools.pack
import net.sergeych.utools.unpack

private val instanceCounter = AtomicCounter()

class DatagramClient<S>(
    private val connector: DatagramConnector,
    val serverAddress: NetworkAddress,
    private val localInterface: KiloInterface<S>,
    private val clientSigningKey: Key.Signing? = null,
    private val createSession: () -> S,
) : LogTag("DGC:${instanceCounter.next()}") {

    private var sessionId = 0u
    private var isClosed = false
    val job = globalLaunch {
        while (isActive && !isClosed) {
            openSession()
            pumpSession()
            debug { "closing session $sessionId" }
            sessionId = 0u
        }
    }

    private suspend fun pumpSession() {
        try {
            val proxyDevice = ProxyDevice()
            val client = KiloClientConnection<S>(localInterface, proxyDevice, createSession(), clientSigningKey)
            coroutineScope {
                fun stop() {
                    cancel()
                }
                launch {
                    while (isActive) {
                        if (sessionId == 0u) {
                            error { "invalid zero sessionId in pump" }
                            stop()
                        } else {
                            val data = proxyDevice.output.receive()
                            if (data == null) {
                                warning { "kiloconnection send empty frame: closing" }
                                stop()
                            } else {
                                send(OuterPacket(sessionId, data))
                            }
                        }
                    }
                }
                while (isActive) {
                    val packet: OuterPacket = unpack(connector.incoming.receive().message)
                    if (packet.id == 0u) {
                        val result: DatagramResult = unpack(packet.payload)
                        if (result is DatagramResult.Error) {
                            if (result.error == DatagramError.Closed) {
                                info { "connection is closed on the remote part" }
                            } else {
                                warning { "unexpected error while pumping: ${result.error}" }
                            }
                        } else {
                            warning { "unexpected result: $result" }
                        }
                        stop()
                    } else {
                        proxyDevice.input.send(packet.payload)
                    }
                }
            }
            proxyDevice.close()
        } catch (x: CancellationException) {
            throw x
        } catch (x: Exception) {
            x.printStackTrace()
        }
    }

    private suspend fun openSession() {
        while (!isClosed) {
            send(OuterPacket(0u, ubyteArrayOf()))
            val packet: OuterPacket = unpack(connector.incoming.receive().message)
            if (packet.id == 0u) {
                when (val result: DatagramResult = unpack(packet.payload)) {
                    is DatagramResult.Error -> {
                        warning { "error opening connection: ${result.error}" }
                    }

                    is DatagramResult.Ok -> {
                        debug { "connection established, our id is ${result.id}" }
                        sessionId = result.id
                        break
                    }
                }
            } else {
                warning { "received unexpected openSession id: ${packet.id}" }
            }
            delay(500)
        }
    }


    private suspend fun send(packet: OuterPacket) {
        connector.send(pack(packet), serverAddress)
    }

    fun close() {
        job.cancel()
    }
}