package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.isActive
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.datetime.Instant
import net.sergeych.crypto.Key
import net.sergeych.kiloparsec.KiloInterface
import net.sergeych.kiloparsec.KiloServerConnection
import net.sergeych.kiloparsec.ProxyDevice
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.debug
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter
import net.sergeych.utools.now
import net.sergeych.utools.pack
import net.sergeych.utools.unpack

private val instanceCounter = AtomicCounter()

private class Connection<S>(
    val server: DatagramServer<S>,
    val connection: KiloServerConnection<S>,
    val device: ProxyDevice,
    val connectionId: ConnectionId,
    var lastActiveAt: Instant = now(),
) {

    suspend fun proceedFrame(frame: OuterPacket) {
        lastActiveAt = now()
        if (frame.payload.isEmpty()) {
            // ping
            server.sendData(connectionId,ubyteArrayOf())
        } else
            device.input.send(frame.payload)
    }

    private val job = globalLaunch {
        while (isActive) {
            val x = device.output.receive() ?: break
            server.sendData(connectionId, x)
        }
    }

    suspend fun close() {
        job.cancel()
        server.remove(connectionId)
    }
}

class DatagramServer<S>(
    private val connector: DatagramConnector,
    private val localInterface: KiloInterface<S>,
    private val serverSigningKey: Key.Signing? = null,
    private val createSession: () -> S,
) : LogTag("DGS:${instanceCounter.next()}") {

    private val job = globalLaunch {
        while (isActive) {
            try {
                val datagram = connector.incoming.receive()
                val outer = unpack<OuterPacket>(datagram.message)
                if (outer.id == 0u)
                    setupNewConnection(datagram.address, outer)
                else
                    proceedExistingConnection(datagram.address, outer)
            } catch (_: ClosedReceiveChannelException) {

            }
        }
    }

    private val access = Mutex()
    private val connections = mutableMapOf<ConnectionId, Connection<S>>()

    private fun findFreeId(address: NetworkAddress): UInt {
        val existingIds = connections.keys.filter { it.address == address }.map { it.id }.toSet()
        debug { "existing connections from $address: $existingIds" }
        var i = 0u
        while (i in existingIds) i++
        debug { "found the free id: $i" }
        return i
    }

    private suspend fun setupNewConnection(address: NetworkAddress, packet: OuterPacket) {
        val id = access.withLock {
            val connectionId = ConnectionId(address, findFreeId(address))
            val proxy = ProxyDevice()
            connections[connectionId] = Connection(
                this,
                KiloServerConnection(localInterface, proxy, createSession(), serverSigningKey),
                proxy,
                connectionId,
            )
            connectionId
        }
        sendResult(id,DatagramResult.Ok(id.id))
    }

    private suspend fun proceedExistingConnection(address: NetworkAddress, outerPacket: OuterPacket) {
        val connectionId = ConnectionId(address, outerPacket.id)
        val c = access.withLock { connections[connectionId] }
        if (c == null) {
            sendResult(connectionId, DatagramResult.Error(DatagramError.NotExists))
        } else {
            c.proceedFrame(outerPacket)
        }
    }

    /**
     * Important. Closing the server also closes the connector which is supposed to be "owned"
     * exclusively by `DatagramServer`
     */
    fun close() {
        job.cancel()
        runCatching {
            connector.close()
        }
    }

    internal suspend fun remove(connectionId: ConnectionId) {
        access.withLock {
            connections.remove(connectionId)
        }

    }

    internal suspend fun sendData(connectionId: ConnectionId,data: UByteArray) {
        connector.send(pack(OuterPacket(connectionId.id, data)), connectionId.address)
    }

    internal suspend fun sendResult(connectionId: ConnectionId, result: DatagramResult) {
        connector.send(pack(OuterPacket(0u, pack(result))), connectionId.address)
    }
}